/**
 * Utility functions for algorithms management.
 */

#include <stdio.h>
#include <xdc/std.h>
#include <ti/xdais/ialg.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>
#include <ti/sysbios/family/c66/Cache.h>

#include "Heaps.h"
#include "AlgorithmUtils.h"

static void Algorithm_free(IALG_MemRec memTabs[], int n);
static ResultCode Algorithm_allocate(IALG_MemRec memTabs[], int n);

#define L2_SCRATCH_SIZE (0x3A000)
#define DDR3_SCRATCH_SIZE (0x400000)

#pragma DATA_SECTION(l2Scratch, ".l2")
static int8_t l2Scratch[L2_SCRATCH_SIZE];

static int8_t* ddr3Scratch = NULL;

IALG_Handle Algorithm_create(IALG_Fxns *fxns, IALG_Params *params, IALG_MemRec *memTabs)
{
    IALG_Handle alg;
    IALG_Fxns *parentFxns = NULL;

    if (fxns == NULL) {
        return NULL;
    }

    int n = (fxns->algNumAlloc != NULL) ? fxns->algNumAlloc() : IALG_DEFMEMRECS;

    if ((n > MAX_MEM_RECS) || (n < 0)) {
        printf("algNumAlloc failed: %d\n", n);
        fflush(stdout);

        return NULL;
    }

    n = fxns->algAlloc(params, &parentFxns, memTabs);

    if (n < 0) {
        puts("algAlloc failed");
        return NULL;
    }

    if (n > MAX_MEM_RECS) {
        puts("Too many memory records required");
        return NULL;
    }

    if (Algorithm_allocate(memTabs, n) != ResultCode_Ok) {
        puts("algorithm_allocate failed");
        fflush(stdout);

        return NULL;
    }

    alg = (IALG_Handle) memTabs[0].base;
    alg->fxns = fxns;

    int ret = fxns->algInit(alg, memTabs, NULL, params);

    if (ret != IALG_EOK) {
        printf("algInit failed: %d\n", ret);
        fflush(stdout);

        //fxns->algFree(alg, memTabs);
        Algorithm_free(memTabs, n);

        return NULL;
    }

    return alg;
}

ResultCode Algorithm_delete(IALG_Handle alg)
{
    IALG_MemRec memTab[MAX_MEM_RECS];
    int n;

    IALG_Fxns *fxns = alg->fxns;

    n = (fxns->algNumAlloc != NULL) ? fxns->algNumAlloc() : IALG_DEFMEMRECS;

    if (n > MAX_MEM_RECS) {
        return ResultCode_Error;
    }

    memTab[0].base = alg;

    n = fxns->algFree(alg, memTab);

    if (n > MAX_MEM_RECS) {
        return ResultCode_Error;
    }

    Algorithm_free(memTab, n);

    return ResultCode_Ok;
}

static inline int isMemSpaceExternal(IALG_MemSpace space)
{
    return ((space & IALG_MXTRN) != 0);
}

static void Algorithm_free(IALG_MemRec memTabs[], int n)
{
    int i;
    IALG_MemRec *memTab;

    for (i = 0; i < n; i++) {
        memTab = &memTabs[i];

        if (memTab->attrs != IALG_SCRATCH) { // Free only non-scratch memory
            Memory_free(isMemSpaceExternal(memTab->space) ? ddr3Heap : NULL, memTab->base, memTab->size);
        }

        memTab->base = NULL;
        memTab->size = 0;
    }
}

static ResultCode Algorithm_allocate(IALG_MemRec memTabs[], int n)
{
    int i;
    IALG_MemRec *memTab;
    Error_Block eb;

    Error_init(&eb);

    // Initialize DDR3 scratch if it is not initialized
    if (ddr3Scratch == NULL) {
        ddr3Scratch = Memory_alloc(ddr3Heap, DDR3_SCRATCH_SIZE, L2_CACHE_LINE_SIZE, &eb);

        if (ddr3Scratch == NULL) {
            puts("Failed to allocate DDR3 scratch memory block");
            return ResultCode_Error;
        }

        //printf("DDR3 scratch %p %p\n", ddr3Scratch, ddr3Scratch + DDR3_SCRATCH_SIZE - 64);
    }

    int32_t l2ScratchPointer = (int32_t) l2Scratch;
    int32_t ddr3ScratchPointer = (int32_t) ddr3Scratch;

    for (i = 0; i < n; i++) {
        memTab = &memTabs[i];

        if (memTab->attrs == IALG_SCRATCH) { // Scratch memory
            // Define on-chip or off-chip scratch
            if (isMemSpaceExternal(memTab->space)) {
                int alignment = (memTab->alignment <= L2_CACHE_LINE_SIZE) ? L2_CACHE_LINE_SIZE : memTab->alignment;

                // Align scratch
                int rest = ddr3ScratchPointer % alignment;

                if (rest) {
                    ddr3ScratchPointer += alignment - rest;
                }

                // Allocate scratch
                memTab->base = (void*) ddr3ScratchPointer;
                ddr3ScratchPointer += memTab->size;

                if (ddr3ScratchPointer - (int32_t)ddr3Scratch >= DDR3_SCRATCH_SIZE) {
                    memTab->base = NULL;
                }
            } else {
                // Align scratch
                int rest = l2ScratchPointer % memTab->alignment;

                if (rest) {
                    l2ScratchPointer += memTab->alignment - rest;
                }

                // Allocate scratch
                memTab->base = (void*) l2ScratchPointer;
                l2ScratchPointer += memTab->size;

                if (l2ScratchPointer - (int32_t)l2Scratch >= L2_SCRATCH_SIZE) {
                    memTab->base = NULL;
                }
            }
        } else {
            if (isMemSpaceExternal(memTab->space)) {
                memTab->base = Memory_alloc(ddr3Heap, memTab->size, (memTab->alignment <= L2_CACHE_LINE_SIZE) ? L2_CACHE_LINE_SIZE : memTab->alignment, &eb);
            } else {
                memTab->base = Memory_alloc(NULL, memTab->size, memTab->alignment, &eb);
            }
        }

        if (memTab->base == NULL) {
            // Memory allocation error - free all previously allocated blocks
            int j;
            for (j = 0; j < i; j++) {
                memTab = &memTabs[j];

                if (memTab->attrs != IALG_SCRATCH) { // Free only non-scratch memory
                    Memory_free(isMemSpaceExternal(memTab->space) ? ddr3Heap : NULL, memTab->base, memTab->size);
                }

                memTab->base = NULL;
                memTab->size = 0;
            }

            return ResultCode_Error;
        }
    }

    return ResultCode_Ok;
}

void Algorithm_synchronize(IALG_MemRec* memTabs)
{
    int i = 0;
    for (; i < MAX_MEM_RECS; i++) {
        IALG_MemRec* memTab = &memTabs[i];

        if (memTab->base == NULL) {
            break;
        }

        if (isMemSpaceExternal(memTab->space)) {
            Cache_wbInv(memTab->base, memTab->size, Cache_Type_ALL, TRUE);
        }
    }
}
