/**
 * Task that decodes.
 */

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <xdc/std.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/family/c66/Cache.h>

#include "H264HighProfileDecoder.h"
#include "BufferDescriptorManager.h"
#include "Heaps.h"
#include "Configuration.h"
#include "GlobalGates.h"
#include "MulticoreApi.h"
#include "DecodeTaskCommon.h"

#define INPUT_DATA_WAIT_INTERVAL (10000 / Clock_tickPeriod) // 10 milliseconds
#define MAX_BUFFER_IDS_TO_FREE (16)

extern cregister volatile unsigned int DNUM;

static const int8_t* const inputBufferStartAddress = (int8_t*) 0x90000000;
static const int8_t* const inputBufferEndAddress = (int8_t*) (0x90000000 + 4 * 1024 * 1024);
static const int8_t* const outputBufferStartAddress = (int8_t*) 0x98000000;

void decodeMasterTask(UArg decoderIndex, UArg arg1)
{
    int i;
    int result;
    Error_Block eb;
    Decoder* decoder = NULL;
    IVIDDEC2_Handle decoderHandle;
    IVIDDEC2_Fxns *decoderFunctions;
    IVIDDEC2_DynamicParams *decoderDynamicParams;
    IVIDDEC2_Status* decoderStatus;
    IVIDDEC2_InArgs* inDecoderArguments;
    IVIDDEC2_OutArgs* outDecoderArguments;
    IALG_MemRec* memTabs;

    int8_t* inputBufferPointer = (int8_t*) inputBufferStartAddress;
    int8_t* outputBufferPointer = (int8_t*) outputBufferStartAddress;

    XDM1_BufDesc inDecoderBufferDescriptor = { 0 };
    XDM_BufDesc *outDecoderBufferDescriptor = NULL;
    int inputDecoderBufferSize;
    int8_t *inputDecoderBuffer = NULL;
    BufferDescriptorManagerHandle decoderBufferManager = NULL;
    int bytesAvaliable = 0;
    const DspConfiguration* configuration = Configuration_get();
    const DspDecoderConfiguration* decoderConfiguration = &configuration->decoderConfigurations[decoderIndex];
    IArg gateKey;

    // Shared data, synchronization, multi-core
    int algorithmIndex = Common_getDecoderAlgorithmIndex(decoderIndex);
    GateMP_Handle sharedDataGate = GlobalGates_getSharedDecoderDataGate();
    int encodersCount = 0;
    SharedDecoderData* sharedDecoderData = &sharedDecodersData[decoderIndex];

    Error_init(&eb);

    printf("Initializing master decoder %d\n", decoderIndex);
    fflush(stdout);

    switch (decoderConfiguration->decoderType) {
    case H264HP_DECODER:
        decoder = H264HighProfileDecoder_create(decoderConfiguration, decoderIndex);
        break;

    default:
        puts("Unknown decoder type");
    }

    if (decoder == NULL) {
        puts("Failed to create decoder");
        goto TASK_END;
    }

    printf("Create decoder %d %dx%d@%d\n", decoderIndex,
            decoderConfiguration->inputWidth,
            decoderConfiguration->inputHeight,
            decoderConfiguration->inputFrameRate);

    decoderHandle = decoder->handle;
    decoderStatus = decoder->status;
    inDecoderArguments = decoder->inArguments;
    outDecoderArguments = decoder->outArguments;
    decoderDynamicParams = decoder->dynamicParams;
    decoderFunctions = decoder->handle->fxns;
    memTabs = decoder->memTabs;

    puts("Decoder created");
    fflush(stdout);

    puts("Configure decoder buffers");

    decoder->activate(decoder);
    result = decoderFunctions->control(decoderHandle, XDM_GETBUFINFO, decoderDynamicParams, decoderStatus);
    decoder->deactivate(decoder);

    if (result != IVIDDEC2_EOK) {
        printf("XDM_GETBUFINFO failed (%d)\n", result);
        goto TASK_END;
    }

    printf("minNumInBufs %d\nminNumOutBufs %d\n", decoderStatus->bufInfo.minNumInBufs, decoderStatus->bufInfo.minNumOutBufs);

    for (i = 0; i < decoderStatus->bufInfo.minNumInBufs; i++) {
        printf("minInBufSize[%d] %d\n", i, decoderStatus->bufInfo.minInBufSize[i]);
    }

    for (i = 0; i < decoderStatus->bufInfo.minNumOutBufs; i++) {
        printf("minOutBufSize[%d] %d\n", i, decoderStatus->bufInfo.minOutBufSize[i]);
    }

    if (decoderStatus->bufInfo.minNumInBufs != 1) {
        printf("Program is designed to work with single input buffer (%d input buffers required)\n",
                decoderStatus->bufInfo.minNumInBufs);
        goto TASK_END;
    }

    if (decoderStatus->bufInfo.minNumOutBufs != 3) {
        puts("Only 3 YUV planes supported");
        goto TASK_END;
    }

    inputDecoderBufferSize = decoderStatus->bufInfo.minInBufSize[0];
    inputDecoderBuffer = Memory_alloc(ddr3Heap, inputDecoderBufferSize, L2_CACHE_LINE_SIZE, &eb);

    if (inputDecoderBuffer == NULL) {
        puts("Failed to allocate input decoder buffer");
        goto TASK_END;
    }

    inDecoderBufferDescriptor.numBufs = 1;
    inDecoderBufferDescriptor.descs[0].bufSize = inputDecoderBufferSize;
    inDecoderBufferDescriptor.descs[0].buf = (XDAS_Int8*) inputDecoderBuffer;

    decoderBufferManager = BufferDescriptorManager_create(512 * (encodersCount + 1), decoderStatus->bufInfo.minNumOutBufs, decoderStatus->bufInfo.minOutBufSize);

    if (decoderBufferManager == NULL) {
        puts("Failed to create buffer manager for decoder");
        goto TASK_END;
    }

    puts("Decoding");
    fflush(stdout);

    inDecoderArguments->inputID = 0;

    while (1) {
        if ((inputBufferPointer < inputBufferEndAddress) && (bytesAvaliable < inputDecoderBufferSize)) {
            int bytesToRead = inputDecoderBufferSize - bytesAvaliable;
            if (bytesToRead > inputBufferEndAddress - inputBufferPointer) {
                bytesToRead = inputBufferEndAddress - inputBufferPointer;
            }

            memcpy(inDecoderBufferDescriptor.descs[0].buf + bytesAvaliable, inputBufferPointer, bytesToRead);

            bytesAvaliable += bytesToRead;
            inputBufferPointer += bytesToRead;
        } else {
            puts("No more data. Test end.");
            printf("You can read output YUV from address 0x%p size %d (size in words = %d)\n", outputBufferStartAddress, outputBufferPointer - outputBufferStartAddress, (outputBufferPointer - outputBufferStartAddress) / 4);
            fflush(stdout);
            goto TASK_END;
        }

        // Prepare input and output structures

        if (outDecoderBufferDescriptor == NULL) { // New output decoder buffer needed
            // ID of the new output buffer
            if (inDecoderArguments->inputID == INT_MAX) { // Handle integer overflow
                inDecoderArguments->inputID = 1;
            } else {
                inDecoderArguments->inputID++;
            }

            outDecoderBufferDescriptor = BufferDescriptorManager_getUnusedBufferDescriptor(decoderBufferManager, inDecoderArguments->inputID);

            if (outDecoderBufferDescriptor == NULL) {
                puts("Failed to create output buffer");
                goto TASK_END;
            }

            Cache_inv(outDecoderBufferDescriptor->bufs[0], outDecoderBufferDescriptor->bufSizes[0], Cache_Type_ALL, TRUE);
            Cache_inv(outDecoderBufferDescriptor->bufs[1], outDecoderBufferDescriptor->bufSizes[1], Cache_Type_ALL, TRUE);
            Cache_inv(outDecoderBufferDescriptor->bufs[2], outDecoderBufferDescriptor->bufSizes[2], Cache_Type_ALL, TRUE);
        }

        inDecoderArguments->numBytes = bytesAvaliable;

        Cache_wbInv(inputDecoderBuffer, inputDecoderBufferSize, Cache_Type_ALL, TRUE);

        // Share data with slave decoders

        gateKey = GateMP_enter(sharedDataGate);
        Cache_inv(sharedDecoderData, sizeof(*sharedDecoderData), Cache_Type_ALL, TRUE);
        sharedDecoderData->inBufferDescriptor = inDecoderBufferDescriptor.descs[0];
        sharedDecoderData->outBuffer_bufSizes[0] = outDecoderBufferDescriptor->bufSizes[0];
        sharedDecoderData->outBuffer_bufSizes[1] = outDecoderBufferDescriptor->bufSizes[1];
        sharedDecoderData->outBuffer_bufSizes[2] = outDecoderBufferDescriptor->bufSizes[2];
        sharedDecoderData->outBuffer_bufs[0] = outDecoderBufferDescriptor->bufs[0];
        sharedDecoderData->outBuffer_bufs[1] = outDecoderBufferDescriptor->bufs[1];
        sharedDecoderData->outBuffer_bufs[2] = outDecoderBufferDescriptor->bufs[2];
        memcpy(&sharedDecoderData->inArguments, inDecoderArguments, inDecoderArguments->size);
        Cache_wb(sharedDecoderData, sizeof(*sharedDecoderData), Cache_Type_ALL, TRUE);
        GateMP_leave(sharedDataGate, gateKey);

        puts("Decoding frame");
        fflush(stdout);

        // Wait for slave
        decoder->swbarr(decoder->algorithmIndex, decoder->decodersCount);

        // Decode

        decoder->activate(decoder);
        result = decoderFunctions->process(decoderHandle, &inDecoderBufferDescriptor, outDecoderBufferDescriptor, inDecoderArguments, outDecoderArguments);
        Algorithm_synchronize(memTabs);
        decoder->deactivate(decoder);

        Cache_wbInv(outDecoderBufferDescriptor->bufs[0], outDecoderBufferDescriptor->bufSizes[0], Cache_Type_ALL, TRUE);
        Cache_wbInv(outDecoderBufferDescriptor->bufs[1], outDecoderBufferDescriptor->bufSizes[1], Cache_Type_ALL, TRUE);
        Cache_wbInv(outDecoderBufferDescriptor->bufs[2], outDecoderBufferDescriptor->bufSizes[2], Cache_Type_ALL, TRUE);

        // Wait for slave
        decoder->swbarr(decoder->algorithmIndex, decoder->decodersCount);

        if (result != IVIDDEC2_EOK) {
            printf("Decoder process error %d\n", result);

            decoder->activate(decoder);
            result = decoderFunctions->control(decoderHandle, XDM_GETSTATUS, decoderDynamicParams, decoderStatus);
            decoder->deactivate(decoder);

            if (result != IVIDDEC2_EOK) {
                puts("Get status failed");
                goto TASK_END;
            }

            int extendedError = decoderStatus->extendedError;

            printf("Extended error 0x%08X - ", extendedError);

            if (XDM_ISFATALERROR(extendedError)) {
                puts("fatal error");
                goto TASK_END;
            } else {
                puts("non-fatal error");
            }

            fflush(stdout);
        } else {
            puts("Frame decoded");
            fflush(stdout);
        }

        if (!outDecoderArguments->outBufsInUseFlag) {
            // Need to supply new output buffers for next "process" call
            outDecoderBufferDescriptor = NULL;
        }

        if (outDecoderArguments->bytesConsumed >= bytesAvaliable) { // Decoder may consume more bytes than available
            bytesAvaliable = 0;
        } else {
            bytesAvaliable -= outDecoderArguments->bytesConsumed;

            memmove(inDecoderBufferDescriptor.descs[0].buf,
                inDecoderBufferDescriptor.descs[0].buf + outDecoderArguments->bytesConsumed,
                bytesAvaliable);
        }

        // Output buffers that are ready

        i = 0;
        while (1) {
            int outputId = outDecoderArguments->outputID[i];

            if (outputId == 0) {
                break;
            }
            const IVIDEO1_BufDesc* displayBuffer = &outDecoderArguments->displayBufs[i];

            int k;
            for (k = 0; k < 3; k++) {
                memcpy(outputBufferPointer, displayBuffer->bufDesc[k].buf, displayBuffer->bufDesc[k].bufSize);
                outputBufferPointer += displayBuffer->bufDesc[k].bufSize;
            }

            i++;
        }

        // Release buffers not needed to decoder

        int* bufferIdToFreePointer = outDecoderArguments->freeBufID;

        while (*bufferIdToFreePointer != 0) {
            if (BufferDescriptorManager_releaseBufferDescriptor(decoderBufferManager, *bufferIdToFreePointer) != 0) {
                puts("Failed to release out decoder buffer");
                goto TASK_END;
            }

            bufferIdToFreePointer++;
        }
    }

TASK_END:

    fflush(stdout);

    if (decoder != NULL) {
        decoder->delete(decoder);
    }

    if (inputDecoderBuffer != NULL) {
        Memory_free(ddr3Heap, inputDecoderBuffer, inputDecoderBufferSize);
    }

    if (decoderBufferManager != NULL) {
        BufferDescriptorManager_delete(decoderBufferManager);
    }

    fflush(stdout);
}
