#include <stdio.h>
#include <string.h>
#include <xdc/std.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/family/c66/Cache.h>

#include "H264HighProfileDecoder.h"
#include "Heaps.h"
#include "Configuration.h"
#include "MulticoreApi.h"
#include "DecodeTaskCommon.h"
#include "Common.h"

extern cregister volatile unsigned int DNUM;

void decodeSlaveTask(UArg decoderIndex, UArg arg1)
{
    int result;
    Error_Block eb;
    Decoder* decoder = NULL;
    const DspConfiguration* configuration = Configuration_get();
    const DspDecoderConfiguration* decoderConfiguration = &configuration->decoderConfigurations[decoderIndex];
    IVIDDEC2_Handle decoderHandle;
    IVIDDEC2_Fxns *decoderFunctions;
    IVIDDEC2_DynamicParams *decoderDynamicParams;
    IVIDDEC2_Status* decoderStatus;
    IVIDDEC2_OutArgs* outDecoderArguments;
    XDM1_BufDesc inDecoderBufferDescriptor = { 0 };
    XDM_BufDesc outDecoderBufferDescriptor = { 0 };
    IALG_MemRec* memTabs;

    // Shared data, synchronization, multi-core
    int algorithmIndex = Common_getDecoderAlgorithmIndex(decoderIndex);
    SharedDecoderData* sharedDecoderData = &sharedDecodersData[decoderIndex];

    Error_init(&eb);

    printf("Initializing slave decoder %d\n", decoderIndex);
    fflush(stdout);

    switch (decoderConfiguration->decoderType) {
#if ENABLE_H264HP_DECODER == 1
    case H264HP_DECODER:
        decoder = H264HighProfileDecoder_create(decoderConfiguration, decoderIndex);
        break;
#endif
    default:
        puts("Unknown slave decoder type");
        goto TASK_END;
    }

    if (decoder == NULL) {
        puts("Failed to create decoder");
        goto TASK_END;
    }

    decoderHandle = decoder->handle;
    decoderStatus = decoder->status;
    decoderDynamicParams = decoder->dynamicParams;
    outDecoderArguments = decoder->outArguments;
    decoderFunctions = decoder->handle->fxns;
    memTabs = decoder->memTabs;

    inDecoderBufferDescriptor.numBufs = 1;
    outDecoderBufferDescriptor.numBufs = 3;
    outDecoderBufferDescriptor.bufs = sharedDecoderData->outBuffer_bufs;
    outDecoderBufferDescriptor.bufSizes = sharedDecoderData->outBuffer_bufSizes;

    while (1) {
        // Wait for master
        decoder->swbarr(decoder->algorithmIndex, decoder->decodersCount);

        // Read shared data from master decoder
        Cache_inv(sharedDecoderData, sizeof(*sharedDecoderData), Cache_Type_ALL, TRUE);

        inDecoderBufferDescriptor.descs[0] = sharedDecoderData->inBufferDescriptor;

        Cache_inv(sharedDecoderData->inBufferDescriptor.buf, sharedDecoderData->inBufferDescriptor.bufSize, Cache_Type_ALL, TRUE);
        Cache_inv(outDecoderBufferDescriptor.bufs[0], outDecoderBufferDescriptor.bufSizes[0], Cache_Type_ALL, TRUE);
        Cache_inv(outDecoderBufferDescriptor.bufs[1], outDecoderBufferDescriptor.bufSizes[1], Cache_Type_ALL, TRUE);
        Cache_inv(outDecoderBufferDescriptor.bufs[2], outDecoderBufferDescriptor.bufSizes[2], Cache_Type_ALL, TRUE);

        decoder->activate(decoder);
        result = decoderFunctions->process(decoderHandle, &inDecoderBufferDescriptor, &outDecoderBufferDescriptor, &sharedDecoderData->inArguments, outDecoderArguments);
        Algorithm_synchronize(memTabs);
        decoder->deactivate(decoder);

        Cache_wb(outDecoderBufferDescriptor.bufs[0], outDecoderBufferDescriptor.bufSizes[0], Cache_Type_ALL, TRUE);
        Cache_wb(outDecoderBufferDescriptor.bufs[1], outDecoderBufferDescriptor.bufSizes[1], Cache_Type_ALL, TRUE);
        Cache_wb(outDecoderBufferDescriptor.bufs[2], outDecoderBufferDescriptor.bufSizes[2], Cache_Type_ALL, TRUE);

        // Wait for master
        decoder->swbarr(decoder->algorithmIndex, decoder->decodersCount);

        if (result != IVIDDEC2_EOK) {
            printf("Decoder process error %d\n", result);

            decoder->activate(decoder);
            result = decoderFunctions->control(decoderHandle, XDM_GETSTATUS, decoderDynamicParams, decoderStatus);
            decoder->deactivate(decoder);

            if (result != IVIDDEC2_EOK) {
                puts("Get status failed");
                goto TASK_END;
            }

            int extendedError = decoderStatus->extendedError;

            printf("Extended error 0x%08X - ", extendedError);

            if (XDM_ISFATALERROR(extendedError)) {
                puts("fatal error");
                goto TASK_END;
            } else {
                puts("non-fatal error");
            }

            fflush(stdout);
        }
    }

TASK_END:

    fflush(stdout);

    if (decoder != NULL) {
        decoder->delete(decoder);
    }
}
